import os
from dataclasses import dataclass

from pants.core.goals.package import BuiltPackage, PackageFieldSet
from pants.core.util_rules.source_files import SourceFiles, SourceFilesRequest
from pants.engine.fs import CreateDigest, Digest, DigestContents, FileContent, MergeDigests
from pants.engine.rules import Get, MultiGet, collect_rules, rule
from pants.engine.target import (
    DependenciesRequest,
    FieldSetsPerTarget,
    FieldSetsPerTargetRequest,
    Target,
    Targets,
)

from pants_backend_kustomize.target_types import (
    KustomizeDependenciesField,
    KustomizeSourcesField,
    KustomizeTarget,
)


@dataclass(frozen=True)
class KustomizationContextRequest:
    target: Target


@dataclass(frozen=True)
class KustomizationContext:
    digest: Digest


@rule
async def prepare_build_context(request: KustomizationContextRequest) -> KustomizationContext:
    root_get = Get(
        SourceFiles,
        SourceFilesRequest([request.target[KustomizeSourcesField]]),
    )

    root_dependencies = await Get(
        Targets,
        DependenciesRequest(request.target[KustomizeDependenciesField]),
    )

    root_contexts_get = []
    other_deps = []
    for dependency in root_dependencies:
        if isinstance(dependency, KustomizeTarget):
            root_contexts_get.append(
                Get(KustomizationContext, KustomizationContextRequest(target=dependency))
            )
        else:
            other_deps.append(dependency)

    embedded_pkgs_per_target_request = Get(
        FieldSetsPerTarget,
        FieldSetsPerTargetRequest(PackageFieldSet, other_deps),
    )

    (root, embedded_pkgs_per_target, *bases) = await MultiGet(
        root_get,
        embedded_pkgs_per_target_request,
        *root_contexts_get,
    )

    # Package binary dependencies for build context.
    embedded_pkgs = await MultiGet(
        Get(BuiltPackage, PackageFieldSet, field_set)
        for field_set in embedded_pkgs_per_target.field_sets
    )

    root_contents = await Get(DigestContents, Digest, root.snapshot.digest)
    root_content = root_contents[0]
    root_dir = os.path.dirname(root_content.path)
    root_manifest = root_content.content.decode()

    embedded_pkgs = embedded_pkgs or []
    if request.target[KustomizeDependenciesField].value:
        for dep, pkg in zip(request.target[KustomizeDependenciesField].value, embedded_pkgs):
            root_manifest = root_manifest.replace(
                dep, os.path.relpath(pkg.artifacts[0].relpath, root_dir)
            )

    patched_root = await Get(
        Digest, CreateDigest([FileContent(root_content.path, root_manifest.encode())])
    )

    input_digest = await Get(
        Digest,
        MergeDigests(
            [
                patched_root,
                root.snapshot.digest,
                *[built_package.digest for built_package in embedded_pkgs],
                *[base.digest for base in bases],
            ]
        ),
    )

    return KustomizationContext(input_digest)


def rules():
    return collect_rules()



from pants.engine.rules import collect_rules

from pants_backend_kustomize import codegen
from pants_backend_kustomize import target_types as targets
from pants_backend_kustomize.util_rules import prepare_context


def target_types():
    return [
        *targets.targets(),
    ]


def rules():
    return [*collect_rules(), *targets.rules(), *codegen.rules(), *prepare_context.rules()]



from __future__ import annotations

from pants.engine.rules import collect_rules
from pants.engine.target import (
    COMMON_TARGET_FIELDS,
    Dependencies,
    MultipleSourcesField,
    SingleSourceField,
    Target,
)
from pants.util.strutil import softwrap


class KustomizeDependenciesField(Dependencies):
    """Dependencies such as packages, secret-decryptors, docker-push-shas, that should be embedded
    into the generated manifest.

    """

    pass


class KustomizeSourceField(SingleSourceField):
    """A file included in a kustomize run. Likely a .yml or .yaml, but could also be other artifacts
    consumed by secret/cm generators."""

    expected_file_extensions = (".yaml", ".yml")
    uses_source_roots = False

    help = softwrap(
        """
        A source file included in a kustomize run. Most likely this'll be other .yaml or .yml files,
        but any other resource that can be put into Kubernetes is valid.
        """
    )


class KustomizeSourcesField(MultipleSourcesField):
    """A set of files feeding into kustomize."""

    alias = "sources"
    uses_source_roots = False

    help = softwrap(
        """
        A group of files used in a kustomize target.
        """
    )


class KustomizeTarget(Target):
    alias = "kustomize"
    core_fields = (
        *COMMON_TARGET_FIELDS,
        KustomizeDependenciesField,
        KustomizeSourcesField,
    )
    help = softwrap(
        """A target for converting a kustomization.yaml and dependencies into a YAML document for
        use with kubectl.

        Example BUILD file:

            kustomize(
                name="kustomize",
                root_src="kustomization.yaml",
                dependencies=[],
                sources=["namespace.yaml", "deployment.yaml", "service.yaml"],
            )
        """
    )


def targets():
    return [
        KustomizeTarget,
    ]


def rules():
    return [
        *collect_rules(),
    ]



from __future__ import annotations

from pants.core.util_rules.external_tool import ExternalTool
from pants.engine.platform import Platform
from pants.option.option_types import BoolOption
from pants.util.strutil import softwrap


class KustomizeTool(ExternalTool):
    options_scope = "kustomize-tool"
    help = "Wrapper for kustomize."

    default_version = "v4.5.7"
    default_known_versions = [
        "v4.5.7|macos_arm64 |3c1e8b95cef4ff6e52d5f4b8c65b8d9d06b75f42d1cb40986c1d67729d82411a|10169842",
        "v4.5.7|macos_x86_64|6fd57e78ed0c06b5bdd82750c5dc6d0f992a7b926d114fe94be46d7a7e32b63a|10496583",
        "v4.5.7|linux_arm64 |65665b39297cc73c13918f05bbe8450d17556f0acd16242a339271e14861df67|5291095",
        "v4.5.7|linux_x86_64|701e3c4bfa14e4c520d481fdf7131f902531bfc002cb5062dcf31263a09c70c9|5757484",
    ]

    def generate_url(self, plat: Platform) -> str:
        platform_mapping = {
            "macos_arm64": "darwin_arm64",
            "macos_x86_64": "darwin_amd64",
            "linux_arm64": "linux_arm64",
            "linux_x86_64": "linux_amd64",
            "linux_arm64": "linux_arm64",
            "linux_x86_64": "linux_amd64",
        }
        plat_str = platform_mapping[plat.value]
        return (
            f"https://github.com/kubernetes-sigs/kustomize/releases/download/kustomize/{self.version}/"
            f"kustomize_{self.version}_{plat_str}.tar.gz"
        )

    def generate_exe(self, _: Platform) -> str:
        return f"./kustomize"



import os

from pants.core.goals.package import BuiltPackage, PackageFieldSet
from pants.core.target_types import FileSourceField
from pants.core.util_rules.external_tool import DownloadedExternalTool, ExternalToolRequest
from pants.core.util_rules.source_files import SourceFiles, SourceFilesRequest
from pants.engine.fs import (
    AddPrefix,
    CreateDigest,
    Digest,
    DigestContents,
    FileContent,
    MergeDigests,
    RemovePrefix,
    Snapshot,
)
from pants.engine.platform import Platform
from pants.engine.process import Process, ProcessResult
from pants.engine.rules import Get, MultiGet, collect_rules, rule
from pants.engine.target import (
    DependenciesRequest,
    FieldSetsPerTarget,
    FieldSetsPerTargetRequest,
    GeneratedSources,
    GenerateSourcesRequest,
    SourcesField,
    Target,
    Targets,
    TransitiveTargets,
    TransitiveTargetsRequest,
)
from pants.engine.unions import UnionRule
from pants.source.source_root import SourceRoot, SourceRootRequest
from pants_backend_k8s.target_types import KubernetesSourceField

from pants_backend_kustomize.subsystem import KustomizeTool
from pants_backend_kustomize.target_types import KustomizeSourcesField
from pants_backend_kustomize.util_rules.prepare_context import (
    KustomizationContext,
    KustomizationContextRequest,
)


class GenerateKubernetesFromKustomizeRequest(GenerateSourcesRequest):
    input = KustomizeSourcesField
    output = KubernetesSourceField


@rule
async def generate_kubernetes_from_kustomize(
    request: GenerateKubernetesFromKustomizeRequest, kustomize: KustomizeTool
) -> GeneratedSources:
    (context, kustomize) = await MultiGet(
        Get(KustomizationContext, KustomizationContextRequest(request.protocol_target)),
        Get(
            DownloadedExternalTool,
            ExternalToolRequest,
            kustomize.get_request(Platform.current),
        ),
    )

    merged_digest = await Get(
        Digest,
        MergeDigests(
            [context.digest, kustomize.digest],
        ),
    )

    root_dir = request.protocol_target.address.spec_path
    output_files = os.path.join(root_dir, f"{request.protocol_target.address.target_name}.yaml")

    result = await Get(
        ProcessResult,
        Process(
            (
                kustomize.exe,
                "build",
                "--load-restrictor",
                "LoadRestrictionsNone",
                "-o",
                output_files,
                root_dir,
            ),
            input_digest=merged_digest,
            description=f"Generating Python sources from {request.protocol_target.address}.",
            output_files=(output_files,),
        ),
    )

    source_root_request = SourceRootRequest.for_target(request.protocol_target)
    source_root = await Get(SourceRoot, SourceRootRequest, source_root_request)
    source_root_restored = (
        await Get(Snapshot, AddPrefix(result.output_digest, source_root.path))
        if source_root.path != "."
        else await Get(Snapshot, Digest, result.output_digest)
    )

    return GeneratedSources(source_root_restored)


def rules():
    return [
        *collect_rules(),
        UnionRule(GenerateSourcesRequest, GenerateKubernetesFromKustomizeRequest),
    ]
